"""
    Describes the experiment layout w.r.t to the root experiment directory
"""

from . import dispatch

REQUIRED_KEYS = [
    'run_dir',
    'experiment_prefix',
    'model_dir',
    'scared_run_files',
    'results_files',
    'data_files',
    'tmp_dir'
]

_DEFAULT_TEMPLATE = {
    'run_dir': '.',
    'experiment_prefix': None,
    'model_dir': 'models',
    'ignore_dirs': ['models/checkpoints'],
    'scared_run_files': 'models/runs',
    'results_files': 'results',
    'data_files': 'data',
    'run_command': 'python ',
    'delete_dir': None, #this will permanently delete files
    'tmp_dir': 'tmp',
    'bin_dir': 'bin_model_log',
    'local_config': 'experiment_config.yaml',
    'project_config': '../project_config.yaml',
    'global_config': None,
    'result_name_fn': lambda config: '{name}_{_id}'.format(name=config['name'], _id=config['experiment_id']),
    'results_metric_fn': lambda results: results['metrics'], 
    'use_mongo': False
}

@dispatch.register('template', '')
def default_template():
    return _DEFAULT_TEMPLATE

def check_template(template):
    for key in REQUIRED_KEYS:
        if key not in template.keys():
            raise RuntimeError(f'Template requires key: {key}')


def get_template():
    tmpl = dispatch.dispatch('template', '')()
    check_template(tmpl)
    return tmpl




from loguru import logger
import sys

_REGISTERED = {}
 

#Register decorator
def register(group, key):
    """
        matches functions based on group and key
    """

    if group not in _REGISTERED:
        _REGISTERED[group] = {}


    def decorator(fun):
        _REGISTERED[group][key] = fun

        return fun

    return decorator

def dispatch(group, key):

    #oneerror: stops repeated eror being printed to terminal
    @logger.catch(reraise=True, onerror=lambda _: sys.exit(1))
    def find():
        return _REGISTERED[group][key]

    return find()



from sacred import Experiment as SacredExperiment
from sacred.observers import FileStorageObserver, MongoObserver

import sys

#from . import manager
#from . import metrics

import inspect
import os

class Experiment(SacredExperiment):
    def __init__(self,name='exp'):
        caller_globals = inspect.stack()[1][0].f_globals
        # sacred used inspect.stack() to see where the main function has been called from. This is annoying when trying to call an experiment from a different class.
        # to get around this we check if the the experiment has been run from the command line. If not then we change the working dir to the experiments directory.
        if caller_globals['__name__'] == '__main__':
            super(Experiment, self).__init__(name)
        else:
            prev_cwd = os.getcwd()
            os.chdir('../')
            super(Experiment, self).__init__(name)
            os.chdir(prev_cwd)

        self.config_function = None

    def configs(self, function):
        self.config_function = function

    def run_config(self, function, config, use_observer=True):
        self.observers = [] #reset observed
        self.configurations = []

        if use_observer:
            self.observers.append(FileStorageObserver('runs'))

        self.add_config(config)
        captured_function = self.main(lambda: function(config))
        self.run(captured_function.__name__)

    def log_metrics(self, X, Y, prediction_fn, var_flag=True, log=True, prefix=None):
        return metrics.log_regression_scalar_metrics(self, X, Y, prediction_fn, var_flag=var_flag, log=log, prefix=prefix)

    def automain(self, function):
        """
        automain, copies the functionality of sacred automain. Will run the experiment using a specific input

        when the file is run from command line we support the follow command line arguments:
            int>=0: run specific config id [default = 0]
            -1: run all experiment configs

        """

        #check if the function was run through the command line or imported
        #if imported we do not run the experiments
        captured = self.main(function)
        if function.__module__ != "__main__":
            return

        if self.config_function is None:
            raise RuntimeError('No config function registered. ')

        filename = inspect.getfile(function)

        use_observer=True
        i = 0
        if len(sys.argv) == 2:
            i = int(sys.argv[1])

        if len(sys.argv) == 3:
            i = int(sys.argv[1])
            flag = int(sys.argv[2])

            use_observer = flag

        configs = self.config_function()

        if i == -1:
            for i, config in enumerate(configs):
                config = manager.ensure_correct_fields_for_model_file_config(filename, config, i)
                self.run_config(function, config, use_observer=use_observer)
                
        else:
            config = manager.ensure_correct_fields_for_model_file_config(filename, configs[i], i)
            self.run_config(function, config, use_observer=use_observer)




import typer

from .. import state
from .. import dispatch
from ..computation import manager, cluster

def sync(location: str = typer.Option("local", help=state.help_texts['location'])):
    #load experiment configs and filter
    experiment_config = state.experiment_config

    #if cluster 
    if location in experiment_config.keys():
        if experiment_config[location]['type'] == 'cluster':
            fn = dispatch.dispatch('sync', 'cluster')
        else:
            #get relevant run function
            fn = dispatch.dispatch('sync', location)

    else:
        #get relevant run function
        fn = dispatch.dispatch('sync', location)

    fn(location)


@dispatch.register('sync', 'local')
def local_sync(location):
    print('sync')
    #local_runner.local_run(configs_to_run, run_settings)

@dispatch.register('sync', 'cluster')
def cluster_sync(location):
    cluster.sync_with_cluster(location)




import typer

from .. import state
from .. import dispatch
from ..computation import manager, local_runner, docker_runner, cluster
from .. import utils

def construct_filter(_filter, filter_file):
    filter_dict = utils.str_to_dict(_filter)

    filter_from_file = {}
    if filter_file is not None:
        #filter from file will overwrite filter_dict
        filter_from_file = utils.json_from_file(filter_file)

    filter_dict =  utils.add_dicts([filter_dict, filter_from_file])
    return filter_dict

def run(
    location: str = typer.Option("local", help=state.help_texts['location']),
    force_all: bool = typer.Option(True, help=state.help_texts['force_all']),
    observer: bool = typer.Option(True, help=state.help_texts['observer']),
    filter: str = typer.Option("{}", help=state.help_texts['filter']),
    filter_file: str = typer.Option(None, help=state.help_texts['filter_file']),
    sbatch: bool = typer.Option(True, help='If true will automatically call sbatch to run files on cluster'),
):

    #construct filter from passed input and file input
    filter_dict = construct_filter(filter, filter_file)
    
    #group together params so passing them around is easier
    run_settings = {
        'observer': observer,
        'force_all': force_all,
        'run_sbatch': sbatch,
    }

    #load experiment configs and filter
    configs_to_run = manager.get_configs_from_model_files()
    configs_to_run = manager.filter_configs(configs_to_run, filter_dict)

    experiment_config = state.experiment_config

    #if cluster 
    if location in experiment_config.keys():
        if experiment_config[location]['type'] == 'cluster':
            fn = dispatch.dispatch('run', 'cluster')
        else:
            #get relevant run function
            fn = dispatch.dispatch('run', location)
    else:
        #get relevant run function
        fn = dispatch.dispatch('run', location)

    fn(configs_to_run, run_settings, location)


@dispatch.register('run', 'local')
def local_run(configs_to_run, run_settings, location):
    local_runner.local_run(configs_to_run, run_settings)

@dispatch.register('run', 'docker')
def docker_run(configs_to_run, run_settings, location):
    docker_runner.docker_run(configs_to_run, run_settings)

@dispatch.register('run', 'cluster')
def cluster_run(configs_to_run, run_settings, location):
    cluster.cluster_run(configs_to_run, run_settings, location)





from pathlib import Path
import shutil
import typing
import os
import yaml
import itertools
import hashlib
import json
import uuid
import copy

import importlib
import importlib.util
from loguru import logger
import typing
import zipfile

YES_LIST = ('yes', 'true', 't', 'y', '1')
NO_LIST = ('no', 'false', 'f', 'n', '0')

def _s(*args) -> str:
    """
        *args: List[str]
    """
    s_arr = [str(s) for s in args]
    return ''.join(s_arr)

def str_to_bool(v: str) -> bool:
    #from https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
    if isinstance(v, bool):
       return v
    if v.lower() in YES_LIST:
        return True
    elif v.lower() in NO_LIST:
        return False
    else:
        raise RuntimeError('Boolean value expected.')

def str_to_dict(s: str) -> dict:
    return json.loads(s)

def json_from_file(f: str) -> dict:
    _dict = {}

    with open(f, 'r') as fh:
        _dict = json.load(fh)

    return _dict

def get_permission(question):
    ans = input(question)
    if ans in YES_LIST:
        return True
    return False

def ask_permission(question, fn=None):
    #Ask permission before running fn
    ans = get_permission(question)

    if fn is None:
        return ans
    else:
        if ans:
            fn()



def move_dir_if_exists(root, tmp_dir):
    if os.path.exists(root):
        shutil.move(root, tmp_dir)

def mkdir_if_not_exists(root):
    Path(root).mkdir(exist_ok=True)

def remove_dir_if_exists(root):
    if os.path.exists(root):
        shutil.rmtree(root)

def delete_if_empty(d):
    if len(os.listdir(d)) == 0:
        remove_dir_if_exists(d)

def split_path(path: str):
    return os.path.normpath(path).split(os.path.sep)

def print_dict(_dict):
    print(json.dumps(_dict, indent=3))

def add_dicts(dict_array: typing.List[dict], deepcopy=False) -> dict:
    """
        Adds all dicts in dict_array into the first dict and returns this.
    """
    sum_dict = None
    for d in dict_array:
        if deepcopy:
            d = copy.deepcopy(d)

        if sum_dict is None:
            sum_dict = d
        else:
            sum_dict.update(d)

    return sum_dict

def dict_is_subset(dict1, dict2):
    """
        Return true if dict1 is a subset of dict2. 
            If dict1 has iterable items then this will be treated as an OR function.
    """
    #return all(item in dict2.items() for item in dict1.items())
    is_subset=True
    for k, i in dict1.items():
        if k not in dict2.keys():
            is_subset=False
            break

        if type(i) is list:
            #if is a list then either the element is a list or we need to sum over the list
            if type(dict2[k]) is list:
                #matching lists
                if dict2[k] == dict1[k]:
                    continue

            else:
                #loop through dict1 and check if any match
                any_matched = False
                for item in dict1[k]:
                    if dict2[k] == item:
                        any_matched = True
                        break
                if not any_matched:
                    is_subset=False
                    break
        else:
            if dict1[k] != dict2[k]:
                is_subset=False
                break

    return is_subset

def get_dict_without_key(_dict, key):
    return {i:_dict[i] for i in _dict if i != key}

def get_dict_hash(_dict):
    """
        Args:
            _dict (dict) - a single configuration
        Returns:
            md5 hash of the sorted _dict
    """
    return hashlib.md5(json.dumps(_dict, sort_keys=True).encode('utf-8')).hexdigest()

def get_unique_key():
    return uuid.uuid4().hex

def load_mod(root, file_name):
    """ 
        loads file_name as a python module
    """
    cwd = os.getcwd()
    #cd into root so that relative paths inside file_name still work
    os.chdir(root)
    spec = importlib.util.spec_from_file_location("", file_name)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    #revert back to orginal working directory
    os.chdir(cwd)
    return foo

def read_yaml(file_name):
    #load experiment config
    with open(file_name) as f:
        # The FullLoader parameter handles the conversion from YAML
        # scalar values to Python the dictionary format
        ec = yaml.load(f, Loader=yaml.FullLoader)

    return ec

def get_all_permutations(options):
    #get all permutations of options
    keys, values = zip(*options.items())
    permutations_dicts = [dict(zip(keys, v)) for v in itertools.product(*values)]
    return permutations_dicts

def zip_dir(path, zipf, ignore_dir_arr=None, dir_path=None):
    """
        Path can be something like ../../folder_1/folder_2/lib/*
        we strip all leading directory structure and zip only lib/*
    """
    #target_dir = os.path.basename(os.path.normpath(path))

    path_split = os.path.normpath(path).split(os.sep)

    if dir_path is None:
        dir_path=''

    for root, dirs, files in os.walk(path):
        for f in files:
            if ignore_dir_arr is not None:
                ignore_flag = False
                for ignore_dir in ignore_dir_arr:
                    #if the root starts with ignore dir then we do not want to zip it
                    if str(root).startswith(ignore_dir):
                        ignore_flag = True
                if ignore_flag:
                    #do not zip this dir
                    continue 

            #remove leading directory structure
            root_split = os.path.normpath(root).split(os.sep)
            if dir_path:
                target_dir = dir_path
            else:
                target_dir = os.path.join(*root_split[(len(path_split)-1):])

            #zipf.write(os.path.join(root, f))    
            #print(os.path.join(target_dir, os.path.join(root, f)))
            zipf.write(os.path.join(root, f), os.path.join(target_dir, f))    

def ensure_backslash(s):
    if s[-1] != '/':
        return s + '/'
    return s



import typer
from loguru import logger

from . import state #global settings
from . import template

from .computation import startup

from .cli import run, dvc, clean, vis, sync, setup, rollback

app = typer.Typer()

#add run command directly to app
app.command()(run.run)
app.command()(clean.clean)
app.command()(sync.sync)
app.command()(setup.setup)
app.command()(rollback.rollback)

dvc_app = typer.Typer()
app.add_typer(dvc.app, name="dvc")

vis_app = typer.Typer()
app.add_typer(vis.app, name="vis")

@app.callback()
def global_state(ctx: typer.Context, verbose: bool = False, dry: bool = False):
    if verbose:
        #logger.info("Will write verbose output")
        state.verbose = True

    if dry:
        #logger.info("Will write verbose output")
        state.dry = True


    #Ensure that model_log is running in the correct folder etc
    #   This is not required if setup is being called and so we simply check that the command is not setup
    if ctx.invoked_subcommand != 'setup':
        pass_flag = startup.check()

        if not pass_flag:
            exit()

    #load config
    config = startup.load_config()
    state.experiment_config = config

    #load any external files specified in the experiment_config
    startup.load_externals()


def main():
    app()



verbose=False
dry=False

help_texts = {
    'location': 'Location to run code e.g docker/local',
    'force_all': 'Force all experiment to re run, otherwise will only run those that have not already completed',
    'filter': '',
    'filter_file': '',
    'observer': 'Run experiment with a sacred observer'
}

experiment_config = {}





from .. import template
from .. import state
from ..utils import ask_permission
from ..computation import startup, manager

import os
from loguru import logger

def setup():
    #ensure we are not already in an experiment
    if startup.check():
        if state.verbose:
            logger.info('Folder already looks like an experiment. Cannot setup a new one!')

        return
    

    #ask permission to create new folders
    cwd = os.getcwd()
    ask_permission(
        f'Setup experiment here {cwd}?',
        manager.create_default_experiment
    )





import setuptools



setuptools.setup(
    name="sdem", 
    version="0.1.0",
    author="O Hamelijnck",
    author_email="ohamelijnck@gmail.com",
    description="Sacred Experiment Manager",
    long_description="",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    scripts=["cli/sdem"]
)



import typer

from .. import state
from .. import dispatch

from ..computation import local_cleaner, cluster

def clean(
    location: str = typer.Option("local", help=state.help_texts['location']),
):

    if location == 'local':
        clean_local(location)
    else:
        experiment_config = state.experiment_config
        location_type = experiment_config[location]['type']
        if location_type == 'cluster':
            clean_cluster(location)


def clean_local(location):
    local_cleaner.clean()

def clean_cluster(location):
    cluster.clean_up_cluster(location)



import typer

from .. import state
from ..computation.sacred_vis import start_omniboard
from ..computation import sacred_manager

app = typer.Typer()

@app.command('omniboard')
def omniboard():
    exp_name = sacred_manager.get_experiment_name()

    start_omniboard(exp_name)




